
# $Id: Shell.pm 179 2006-10-23 08:36:31Z manu $

package PARALOOP::PLUGIN::Shell;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::PLUGIN::Shell - Shell object

=head1 Derives from...

PARALOOP::PLUGIN::LnInput

=head1 DESCRIPTION

This plugin considers the input file as a list of shell commands: each line
is a little script, executed at a given iteration of MainLoop.

The line may have two formats:

 1/ <some shell code>
 2/ <path_to_an_alternative_interpeter>:<some_code>

When the shell interpreter is not specified (1st format), the shell declared in the
parameter PARALOOP_Shell_interpreter is used. The default value
is /bin/sh. 

when the shell interpreter is specified (2nd format), it is used to execute the rest of the line.

=head2 Examples

The following is a little input file:
 
 echo -n "Hello Folks, Here is my name: "; hostname
 /bin/perl : print "Hello Folks, Here is my name: " . `hostname\n`;
 grep HUMAN file.fasta >file.results 

The first 2 lines write something to the Paraloop.pl output file, 
while the last line writes something to the file called "file.results"

The first and last lines are executed with /bin/sh, the second is executed with perl

=head1 CONFIGURATION FILE PARAMETERS

 PARALOOP_Shell_interpreter 
     path to the default shell interpreter (default /bin/sh)

=head1 PUBLIC METHODS

 WhatParaloopPlugin
     my $msg=$obj->WhatParaloopPlugin($verbosity);

 Parameters
     my $msg=$obj->Parameters();

 New 
     my $obj = PARALOOP::Dummy->New(-params => $params, -log => $o_log);

 Exec
     my ($incr_size,$error,$killed) = $obj->Exec();

 SetOutputFile
     $obj->SetOutputFile('filename');

=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use Cwd;
use File::Basename;
use Carp;

use Logger;
use PARALOOP::PLUGIN::LnInput;
use Runner;
use IO::File;

our @ISA = qw( PARALOOP::PLUGIN::LnInput );

use constant INTERPRETER_DEFAULT => '/bin/sh';

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 179 $ =~ /\d+/g); $r[0]};
}

=pod

=head2 Function WhatParaloopPlugin

 Usage:    my $msg = PARALOOP::Dummy::WhatParaloopPlugin($verbosity)
        or my $msg = $obj->WhatParaloopPlugin()
 Function: Return some information (2 lines max) on the plugin
           An object is NOT considered as a plugin if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopPlugin
{
    my $rvl = "";
    $rvl =
      "PARALOOP::PLUGIN::Shell v$PARALOOP::PLUGIN::Shell::VERSION - Using Runner v$Runner::VERSION - Derivating from PARALOOP::PLUGIN::LnInput v$PARALOOP::PLUGIN::LnInput::VERSION\n";
    $rvl .= "This plugin considers each line of input file as a shell script, and executes it.\n";
    return $rvl;
}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::Dummy::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this plugin
           If your plugin defines some parameters, it is good pratice adding this function,
           to display those parameters when the --parameters switch of paraloop.pl is used.
           However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub Parameters
{
    my $rvl = "";
    $rvl = "\nParameters for the Shell plugin\n=================\n";
    $rvl .= "PARALOOP_Shell_interpreter\tpath to the shell interpreter (default /bin/sh)\n";
    return $rvl;
}

=head2 procedure _Init

 Usage:     CALLED BY UNIVERSAL::New
            my $obj = PARALOOP::Dummy->New(-params => $params, -log => $o_log);
 Access:    private (use New instead)
 procedure: Init the object
 Return:    nothing
 Args:       -params = an object of class ParamParser (already initialized)
             -log    = an object of class Logger (already initialized)
 globals:   none

=cut

sub _Init
{
    my $self    = shift;
    my %args    = @_;
    my $o_param = $args{'-params'};
    my $o_log   = $args{'-log'};

    my ($inputfile, $outputfile);

    if ($o_param->IsDefined('input'))
    {    # check if inputfile is correctly defined
        $inputfile = $o_param->Get('input');
        -r $inputfile or croak("ERROR - $inputfile cannot be opened in read mode");
    }
    else
    {
        croak("ERROR - the input file is not defined");
        $o_param->Usage();
    }

    if ($o_param->IsDefined('output'))
    {    # check if outputfile is correctly defined
        $outputfile = $o_param->Get('output');
        open(TEST, ">>$outputfile") or croak("ERROR - $outputfile cannot be opened in write mode");
        close(TEST);
    }
    else
    {
        croak("ERROR - the output file is not defined");
        &$o_param->Usage();
    }

    my $bin = $o_param->IsDefined('Shell_interpreter') ? $o_param->Get('Shell_interpreter') : INTERPRETER_DEFAULT;
    system("which $bin  >/dev/null 2>&1")==0 or croak "ERROR - $bin cannot be found or executed";

    # Initialize the superclass
    $self->SUPER::_Init(-params => $o_param, -inputfile => $inputfile, -log => $o_log);    # Init the superclass

    # Compute job_id
    my $job_id = $self->_JobId($o_param);

    $self->{'__params'}     = $o_param;
    $self->{'__outputfile'} = $outputfile;
    $self->{'__job_id'}     = $job_id;
    $self->{'__bin'}        = $bin;

    # create a temporary directory
    my $tmp    = $self->{'__tmp'};
    my $stamp  = $o_log->CalcStamp();
    my $tmpdir = $tmp . '/' . &basename($inputfile) . '_' . $job_id . '_' . $stamp;
    mkdir($tmpdir) or croak "ERROR - Cannot create directory $tmpdir: $!";
    $self->{'__tmpdir'} = $tmpdir;

}

=pod

=head2 procedure DESTROY

 Usage:     The destructor of the object
            undef($obj);
 Access:    public
 procedure: Destroyes the resources allocated by _Init
 Return:    nothing
 Args:      nothing
 globals:   none

=cut

sub DESTROY
{
    my $self   = shift;
    my $tmpdir = $self->{'__tmpdir'};
    `rm -rf $tmpdir` if (defined $tmpdir);
}

=pod

=head2 Procedure SetOutputFile

 Usage:     &$opt->SetOutputFile('filename');
 Procedure: change the name of the output file
 Arg :      the output file name

=cut

sub SetOutputFile
{
    my $self       = shift;
    my $outputfile = shift;
    $self->{'__outputfile'} = $outputfile;
}

=pod

=head2 Function Exec

 Usage:    my ($incr_size, $error,$killed) = &$opt->Exec($timeout);
 Function: Read and execute the next line, using the default shell or the specific shell
 Arg:      $timeout   = The number of seconds after which a timeout is sent
 Return:   $incr_size = The number of bytes added to the output file
           $error     = An error code. 
		                If != 0, paraloop WILL BE TERMINATED - Generally marks the end of file
           $killed    = A kill code
	                0: everything OK 
                       -1: killed by timeout
                       -2: killed by TERM signal (kill, qdel, ...)

=cut

sub Exec
{
    my $self    = shift;
    my $timeout = shift;

    my $input_file  = $self->{'__inputfile'};
    my $output_file = $self->{'__outputfile'};
    my $o_log       = $self->{'__log'};
    my $shell_bin   = $self->{'__bin'};
    my $job_id      = $self->{'__job_id'};
    my $tmpdir      = $self->{'__tmpdir'};
    my $rewrite_input =  $self->{'__rewrite_input'};

    my ($nb, $nl) = (0, 0);

    # Read the next line

    my $line_nb   = $self->Tell();
    my $exec_line = $self->NextRecord();

    if (!defined($exec_line))
    {
        my $msg = "NO MORE LINES TO READ FROM $input_file\n";
        $o_log->Trace("$msg", '*', '1');
        return (0, -1, 0);
    }

    # Remember where we are and change to temporary directory

    my $savdir = cwd;    # Go to temporary directory
    chdir($tmpdir) or die "Cannot change to temporary directory";

    # Write a little shell to the temporary directory

    my $etmp = "./exec.shell";
    my $etmpfh = new IO::File(">$etmp") or croak "ERROR - Cannot open $etmp for writing";
    if ($exec_line =~ /\A *(\S+) *:(.+)\Z/)
    {
        print($etmpfh "#! $1\n$2\n");
    }
    else
    {
        print($etmpfh "#! $shell_bin\n$exec_line\n");
    }
    undef($etmpfh);    # close the file

    # Make this shell executable

    chmod(0755, $etmp);

    my $otmp = "exec.out";

    $o_log->Trace("TIMEOUT $timeout\n", '.', 2);

    # Start a timer

    $o_log->StartTimer('Exec_timer');

    # Run the temporary shell

    my ($sts, $killed) = RunExt(-cmd => $etmp, -timeout => $timeout, -stdout => ">$otmp");

    $o_log->Trace("Shell line $line_nb - Return $sts", 'Exec_timer', '1');
    $o_log->Trace(" - Text \"$exec_line\"",            '',           '2');
    $o_log->Trace("\n",                                '',           '1');

    # Remove the temporary shell

    unlink($etmp);

    # Compute the number of bytes that will be added to the output file

    my $rvl_wc = `wc -c $otmp`;
    chomp($rvl_wc);
    ($nb) = ($rvl_wc =~ /\A\s*(\d+)\s*/);

    # Append the output to the main output file

    chdir($savdir);    # Back to work directory
    `cat $tmpdir/$otmp >>$output_file`;

    # If status != 0 and if __rewrite_input on, write the input line to some file
    if ($sts != 0 and $rewrite_input != 0)
    {
        my $new_input_file_name = $output_file.'.in';
        my $new_input = new IO::File(">>$new_input_file_name") or die "Cannot open $new_input_file_name".'.in'." for writing";
        print $new_input "$exec_line\n";
        my $msg = "Input data copied back to " . basename($new_input_file_name);
        $o_log->Trace("$msg\n", '.', 'R');
    };
        
    return ($nb, 0, $killed);
}

