
# $Id: System.pm 162 2006-02-08 08:13:26Z manu $

package PARALOOP::SCHEDULER::System;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::SCHEDULER::System - Scheduler object for SMP machines, or for any architecture with --local

=head1 Derives from...

PARALOOP::SCHEDULER::Scheduler

=head1 DESCRIPTION

This scheduler is used for a multiprocessor SMP machine. It is also used with any architecture, if the --local switch 
is specified.

=head1 CONFIGURATION FILE PARAMETERS

 PARALOOP_Scheduler
      System
 PARALOOP_error_directory 
      The name of the error directory

=head1 PUBLIC METHODS

 WhatParaloopScheduler
    my $msg=$obj->WhatParaloopScheduler()

 Parameters
    my $msg=$obj->Parameters()

 New
    my $obj = PARALOOP::Dummy->New(-params => $params, -log => $log);

 Submit
    my $job_id = $obj->Submit();

 RemTime
    my $rem_time = $obj->RemTime(); 
    
 IsSelfPossible
    If (&$obj->IsSelPossible()==1)...

 JobId
    my $job_id = $obj->JobId();

=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use Cwd;
use File::Basename;
use Carp;

#use Logger;
use Runner;
use PARALOOP::SCHEDULER::Scheduler;

@PARALOOP::SCHEDULER::System::ISA = qw( PARALOOP::SCHEDULER::Scheduler );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 162 $ =~ /\d+/g); $r[0]};

}

=pod

=head2 Function WhatParaloopScheduler

 Usage:    my $msg = PARALOOP::SCHEDULER::System::WhatParaloopScheduler()
        or my $msg = $obj->WhatParaloopScheduler()
 Function: Return some information (2 lines max) on the scheduler
           An object is NOT considered as a scheduler if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopScheduler
{
    my $rvl = "";
    $rvl =
      "PARALOOP::SCHEDULER::System v$PARALOOP::SCHEDULER::System::VERSION - Deriving from Scheduler v$PARALOOP::SCHEDULER::Scheduler::VERSION - Using Runner v$Runner::VERSION\n";
    $rvl .= "This scheduler object is used when the --noqueue switch is provided\n";
    return $rvl;
}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::SCHEDULER::System::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this scheduler
           If your scheduler defines some parameters, it is good pratice adding this function,
           to display those parameters when the --parameters switch of paraloop.pl is used.
	   However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the scheduler parameters (1 line / parameter)

=cut

sub Parameters
{
    my $rvl = "";
    $rvl = "\nParameters for the System scheduler\n=================\n";
    $rvl .= "PARALOOP_Scheduler\tMust be set to System for using this scheduler (default value)\n";
    return $rvl;
}

=pod

=head2 procedure IsSelfPossible

 Usage:  if ($obj->IsSelfPossible()) ... 
 Access: public (use New instead)
 function: return 1 ONLY if slave job: there is no self submit, but we do not
           generate an error. This is important for supporting fair_time_limit in slave mode
 Return:   0/1
 Args:     none
 globals:  none

=cut

sub IsSelfPossible {
    my $self = shift;
    my $o_param = $self->{'__params'};
    if ($o_param->IsDefined('current_mode') and $o_param->Get('current_mode') eq 'SLAVE')
    {
        return 1;
    }
    else
    {
        return 0;
    };
};

=head2 procedure _Init

 Usage:  CALLED BY UNIVERSAL::New
         my $obj = PARALOOP::Dummy->New(-params => $params, -log => $log);
 Access: private (use New instead)
 procedure: Init the object
 Return:   nothing
 Args:     -params = an object of class ParamParser (already initialized)
           -log    = an object of class Logger (already initialized)
 globals:  none

=cut

sub _Init
{
    my $self = shift;
    $self->SUPER::_Init(@_);
    my $o_param = $self->{'__params'};
    my $o_log   = $self->{'__log'};

    #
    # If we have a subcount, we redirect the Io
    #

    if ($o_param->IsDefined('subcount') and ($o_param->Get('subcount')) >= 1)
    {
        $self->IoRedirect();
        return;
    }
}

=pod

=head2 Function Submit

 Usage:     &$obj->Submit()
 Function:  Run another instance of paraloop
 Arg :      none
 Return:    the JobId
=cut

sub Submit
{
    my $self    = shift;
    my $o_param = $self->{'__params'};
    my $o_log   = $self->{'__log'};
    my $subcount = $self->_IncSubCount();
    my $current_mode = $o_param->Get('current_mode');
    my $job_id;
    if ($subcount > 1)
    {
        if ($current_mode eq 'SLAVE')
        {
            $o_log->Trace("SLAVE MODE - THIS JOB WILL BE RESTARTED BY THE MASTER\n", '*', '1');
            $job_id = 0;
        }
        else
        {
            $o_log->Trace("ERROR - THIS SCHEDULER CANNOT RESUBMIT ITSEF\n", '*', '1');
            croak(
                  'ERROR - With this scheduler, paraloop should resubmit itself ONLY 1 time.'
                  );
        }
    }
    else
    {
        print "calling $0 &\n" if ($o_param->IsDefined('verbose'));
        $o_param->Dump('ENV');
        $job_id = &RunExtNoWait(-cmd => $0);
        $o_log->Trace("$0\n", '', '2');
    }
    $self->_SyncLockFile($job_id);
    return $job_id;
}

=pod

=head2 Procedure IoRedirect

 Usage:     &$obj->IoRedirect()
 Procedure: Redirect stdout and stderr to some file ($job_id.EO)
            We DO NOT close STDIN, because it will be impossible later to redirect STDOUT
	    (I think there is due to a bug in perl...)
            Close stdin
 Arg :      none
 Return:    none

=cut

sub IoRedirect
{
    my $self            = shift;
    my $o_param         = $self->{'__params'};
    my $error_directory = $o_param->Get('error_directory');
    my $job_id          = $self->JobId();
    my $stdoe           = $error_directory . '/' . $job_id . '.EO';

    #    close(STDIN);
    close(STDOUT);
    close(STDERR);
    open(STDOUT, "> $stdoe")  or croak("ERROR - Cannot open output file");
    open(STDERR, ">> $stdoe") or croak("ERROR - Cannot open error file");
}

=pod

=head2 Function JobId

 Usage:     my $job_id = $obj->JobId()
 Function:  Ask the system for the job id
            The variable $$ is returned
 Arg :      none
 Return :   the job Id

=cut

sub JobId
{
    my $j = $$ . '.' . `hostname`;
    chomp($j);
    return $j;
}

=pod

=head2 Function IsLaunched

 Usage:    If (&$obj->IsLaunched($job_id)==1)...
 Function: Check the presence of the job in the system
 Arg:      none
 Return:   0 ==> The job is not known by the system
           1 ==> The job is known (may be in different states: Run, Queue, etc)

=cut

sub IsLaunched
{
    my $self    = shift;
    my $o_param = $self->{'__params'};

    my $job_id = shift;
    my $ps_cmd = '';
    if ($job_id =~ /(\d+)\.([a-z0-9_]+)/)    # Ex 3456.medicago
    {
        my $node = $2;
        $job_id = $1;
    }
    $ps_cmd = "ps ax | egrep '^ *" . $job_id . "' | grep -v ' Z' | wc -l";
    print "calling $ps_cmd\n" if ($o_param->IsDefined('verbose'));
    my $rvl = `$ps_cmd`;
    chomp($rvl);
    return $rvl+0;
}

=pod

=head2 Procedure KillJob

 Usage:    $obj->KillJob($job_id);
 Function: Kill the job passed by parameter
           Call kill $job_id
 Arg:      job_id: the job Id to kill
 Return:   none

=cut

sub KillJob
{
    my $self         = shift;
    my $o_param = $self->{'__params'};

    my $job_id       = shift;
    my $kill_cmd = "kill $job_id";
    print "calling $kill_cmd\n" if ($o_param->IsDefined('verbose'));   
    `kill $job_id`;
}

1;

